/*
 * based on the:
 * rotary_encoder.c, (c) 2009 Daniel Mack <daniel@caiaq.de>
 * and
 * gpio_keys.c, Copyright 2005 Phil Blundell
 *
 * state machine code inspired by code from Tim Ruetz
 *
 * A generic driver for rotary encoders connected to GPIO lines.
 * See file:Documentation/input/rotary_encoder.txt for more information
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/gpio.h>
#include <linux/slab.h>
//#include <linux/jiffies.h>// for click debouncing
#include <mach/gpio.h>// at91 specific

#define DRV_NAME "rotary-encoder"

#define GPIO_ROTARY_A AT91_PIN_PC15
#define GPIO_ROTARY_B AT91_PIN_PB10
#define GPIO_ROTARY_S AT91_PIN_PB9

#define CW	BTN_FORWARD
#define CCW	BTN_BACK
#define BTN	BTN_MOUSE

#define INVERTED_A 0
#define INVERTED_B 0
#define REPEAT 0
#define DEBOUNCE_INTERVAL_MS 10

struct rotary_encoder_button_data {
	unsigned int irq;
	struct timer_list timer;
	struct work_struct work;
};

struct rotary_encoder {
	struct input_dev *input;
	struct rotary_encoder_platform_data *pdata;

	unsigned int irq_a;
	unsigned int irq_b;

	bool armed;
	unsigned char dir;	/* 0 - clockwise, 1 - CCW */
	
	struct rotary_encoder_button_data bdata;
};

static struct rotary_encoder* encoder;

//static unsigned long button_time;// in jiffies for click debouncing

static irqreturn_t rotary_encoder_irq(int irq, void *dev_id)
{
	struct rotary_encoder *encoder = dev_id;
#if INVERTED_A
	int a = !gpio_get_value(GPIO_ROTARY_A);
#else
	int a = !!gpio_get_value(GPIO_ROTARY_A);
#endif
#if INVERTED_B
	int b = !gpio_get_value(GPIO_ROTARY_B);
#else
	int b = !!gpio_get_value(GPIO_ROTARY_B);
#endif
	int state;

	state = (a << 1) | b;

	switch (state) {

	case 0x0:
		if (!encoder->armed)
			break;

		input_report_key(encoder->input, encoder->dir ? CCW : CW, 1);// press
		input_sync(encoder->input);
		input_report_key(encoder->input, encoder->dir ? CCW : CW, 0);// release
		input_sync(encoder->input);

		encoder->armed = false;
		break;

	case 0x1:
	case 0x2:
		if (encoder->armed)
			encoder->dir = state - 1;
		break;

	case 0x3:
		encoder->armed = true;
		break;
	}

	return IRQ_HANDLED;
}

static void rotary_encoder_report_event(struct rotary_encoder *encoder)
{
	struct input_dev *input = encoder->input;
	int state = !gpio_get_value(GPIO_ROTARY_S);

	//input_event(input, EV_KEY, KEY_ENTER, state);
	input_report_key(input, BTN, state);
	input_sync(input);
}

static void rotary_encoder_work_func(struct work_struct *work)
{
	struct rotary_encoder *encoder =
		container_of(work, struct rotary_encoder, bdata.work);

	rotary_encoder_report_event(encoder);
}

static void rotary_encoder_timer(unsigned long _data)
{
	struct rotary_encoder *encoder = (struct rotary_encoder *)_data;

	schedule_work(&encoder->bdata.work);
}

static irqreturn_t rotary_encoder_key_irq(int irq, void *dev_id)
{
	struct rotary_encoder *encoder = dev_id;
	struct rotary_encoder_button_data *bdata = &encoder->bdata;

	if (DEBOUNCE_INTERVAL_MS)
		mod_timer(&bdata->timer,
			jiffies + msecs_to_jiffies(DEBOUNCE_INTERVAL_MS));
	else
		schedule_work(&bdata->work);

	return IRQ_HANDLED;
}
/*
this function generate only key click
static irqreturn_t rotary_encoder_key_irq(int irq, void *dev_id)
{
	struct rotary_encoder *encoder = dev_id;
	struct rotary_encoder_platform_data *pdata = encoder->pdata;
	if( time_after(jiffies, button_time + HZ/10) )
	{
		//int s = !gpio_get_value(pdata->gpio_s);
		if( !gpio_get_value(pdata->gpio_s) )
		{
			input_report_key(encoder->input, KEY_ENTER, 1);
			input_report_key(encoder->input, KEY_LEFTSHIFT, 1);
			input_sync(encoder->input);
			input_report_key(encoder->input, KEY_ENTER, 0);
			input_report_key(encoder->input, KEY_LEFTSHIFT, 0);
			input_sync(encoder->input);
		}
	}
	button_time = jiffies;

	return IRQ_HANDLED;
}
*/
static int __init rotary_encoder_init(void)
{
	struct input_dev *input;
	int err;
	//button_time = jiffies;// click debouncing

	encoder = kzalloc(sizeof(struct rotary_encoder), GFP_KERNEL);
	input = input_allocate_device();
	if (!encoder || !input) {
		printk("failed to allocate memory for device\r\n");
		err = -ENOMEM;
		goto exit_free_mem;
	}

	encoder->input = input;
	encoder->irq_a = gpio_to_irq(GPIO_ROTARY_A);
	encoder->irq_b = gpio_to_irq(GPIO_ROTARY_B);
	encoder->bdata.irq = gpio_to_irq(GPIO_ROTARY_S);

	/* create and register the input driver */
	input->name = DRV_NAME;

	/* Enable auto repeat feature of Linux input subsystem */
	if (REPEAT)
		set_bit(EV_REP, input->evbit);

	//input_set_capability(input, EV_KEY, bdata->button.code);
	set_bit(EV_KEY, input->evbit);
	set_bit(CW, input->keybit);
	set_bit(CCW, input->keybit);
	set_bit(BTN, input->keybit);

	setup_timer(&encoder->bdata.timer, rotary_encoder_timer, (unsigned long)encoder);
	INIT_WORK(&encoder->bdata.work, rotary_encoder_work_func);

	err = input_register_device(input);
	if (err) {
		printk("failed to register input device\r\n");
		goto exit_free_mem;
	}

	/* request the GPIOs */
	err = gpio_request(GPIO_ROTARY_A, DRV_NAME);
	if (err) {
		printk("unable to request GPIO %d\r\n",
			GPIO_ROTARY_A);
		goto exit_unregister_input;
	}

	err = gpio_direction_input(GPIO_ROTARY_A);
	if (err) {
		printk("unable to set GPIO %d for input\r\n",
			GPIO_ROTARY_A);
		goto exit_unregister_input;
	}

	err = gpio_request(GPIO_ROTARY_B, DRV_NAME);
	if (err) {
		printk("unable to request GPIO %d\r\n",
			GPIO_ROTARY_B);
		goto exit_free_gpio_a;
	}

	err = gpio_direction_input(GPIO_ROTARY_B);
	if (err) {
		printk("unable to set GPIO %d for input\r\n",
			GPIO_ROTARY_B);
		goto exit_free_gpio_a;
	}

	err = gpio_request(GPIO_ROTARY_S, DRV_NAME);
	if (err) {
		printk("unable to request GPIO %d\r\n",
			GPIO_ROTARY_S);
		goto exit_free_gpio_b;
	}

	err = gpio_direction_input(GPIO_ROTARY_S);
	if (err) {
		printk("unable to set GPIO %d for input\r\n",
			GPIO_ROTARY_S);
		goto exit_free_gpio_b;
	}

	/* request the IRQs */
	err = request_irq(encoder->irq_a, &rotary_encoder_irq,
			  IORESOURCE_IRQ_HIGHEDGE | IORESOURCE_IRQ_LOWEDGE,
			  DRV_NAME, encoder);
	if (err) {
		printk("unable to request IRQ %d\r\n",
			encoder->irq_a);
		goto exit_free_gpio_s;
	}

	err = request_irq(encoder->irq_b, &rotary_encoder_irq,
			  IORESOURCE_IRQ_HIGHEDGE | IORESOURCE_IRQ_LOWEDGE,
			  DRV_NAME, encoder);
	if (err) {
		printk("unable to request IRQ %d\r\n",
			encoder->irq_b);
		goto exit_free_irq_a;
	}

	err = request_irq(encoder->bdata.irq, &rotary_encoder_key_irq,
			  IORESOURCE_IRQ_HIGHEDGE | IORESOURCE_IRQ_LOWEDGE,
			  DRV_NAME, encoder);
	if (err) {
		printk("unable to request IRQ %d\r\n",
			encoder->bdata.irq);
		goto exit_free_irq_b;
	}
	return 0;
exit_free_irq_b:
	free_irq(encoder->irq_b, encoder);
exit_free_irq_a:
	free_irq(encoder->irq_a, encoder);
exit_free_gpio_s:
	gpio_free(GPIO_ROTARY_S);
exit_free_gpio_b:
	gpio_free(GPIO_ROTARY_B);
exit_free_gpio_a:
	gpio_free(GPIO_ROTARY_A);
exit_unregister_input:
	input_unregister_device(input);
	input = NULL; /* so we don't try to free it */
exit_free_mem:
	input_free_device(input);
	kfree(encoder);
	return err;
}

static void __exit rotary_encoder_exit(void)
{
	free_irq(encoder->irq_a, encoder);
	free_irq(encoder->irq_b, encoder);
	free_irq(encoder->bdata.irq, encoder);
	if (DEBOUNCE_INTERVAL_MS)
		del_timer_sync(&encoder->bdata.timer);
	cancel_work_sync(&encoder->bdata.work);
	gpio_free(GPIO_ROTARY_A);
	gpio_free(GPIO_ROTARY_B);
	gpio_free(GPIO_ROTARY_S);
	input_unregister_device(encoder->input);
	kfree(encoder);
}

module_init(rotary_encoder_init);
module_exit(rotary_encoder_exit);

MODULE_ALIAS("platform:" DRV_NAME);
MODULE_DESCRIPTION("GPIO rotary encoder driver");
MODULE_LICENSE("GPL v2");
